/*
 * Copyright (c) 2009 by Stefan Siegl <stesie@brokenpipe.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * For more information on the GPL, please go to:
 * http://www.gnu.org/copyleft/gpl.html
 */

#include <avr/pgmspace.h>
#include <string.h>

#include "config.h"
#include "hardware/storage/sd_reader/sd_raw.h"
#include "hardware/storage/sd_reader/fat.h"
#include "core/vfs/vfs.h"
#include "hardware/storage/sd_reader/vfs_sd.h"
#include "core/debug.h"

#include "protocols/ecmd/ecmd-base.h"


#define MAX_CURRENT_PATH_LEN 80

extern char uploadpath[];

static struct fat_dir_struct *wd;
static struct fat_file_struct *fh;
static char current_path[MAX_CURRENT_PATH_LEN];


void wd_init(void)
{
  current_path[0] = '/';
  current_path[1] = 0;
  wd = vfs_sd_chdir(current_path);
  fh = NULL;

  uploadpath[0] = '/';
  uploadpath[1] = 0;
}


void wd_free(void)
{
  current_path[0] = 0;
  if (fh) fat_close_file(fh);
  if (wd) fat_close_dir(wd);
}


int16_t
parse_cmd_sd_dir (char *cmd, char *output, uint16_t len)
{
  uint16_t year;
  uint8_t month, day, hour, min, sec;

  if (vfs_sd_rootnode == NULL || wd == NULL)
    return ECMD_FINAL(snprintf_P(output, len, PSTR("SD/MMC backend not available.")));

  if (cmd[0] != 0x05) {
    fat_reset_dir(wd);
    cmd[0] = 0x05;
  }

  struct fat_dir_entry_struct dir_entry;
  if (! fat_read_dir(wd, &dir_entry))
    return ECMD_FINAL_OK;

  fat_get_file_modification_date (&dir_entry, &year, &month, &day);
  fat_get_file_modification_time (&dir_entry, &hour, &min, &sec);

  return ECMD_AGAIN(snprintf_P(output, len, PSTR("%10ld  %4d-%02d-%02d  %02d:%02d  %s%c"),
			       dir_entry.file_size,
			       year,month,day,hour,min,
			       dir_entry.long_name,
			       dir_entry.attributes & FAT_ATTRIB_DIR ? '/' : ' '));
}


int16_t
parse_cmd_sd_mkdir (char *cmd, char *output, uint16_t len)
{
  while (*cmd == ' ')
    cmd ++;

  return -vfs_sd_mkdir_recursive (cmd);
}


int16_t
parse_cmd_sd_pwd (char *cmd, char *output, uint16_t len)
{
  int str_length;

  strncpy(output, current_path, len);
  str_length = strlen(output);
  if (str_length > 1) {
    output[str_length] = 0;
    str_length--;
  }
  return ECMD_FINAL(str_length);
}


int16_t
parse_cmd_sd_pup (char *cmd, char *output, uint16_t len)
{
  int str_length;

  strncpy(output, uploadpath, len);
  str_length = strlen(output);
  if (str_length > 1) {
    output[str_length] = 0;
    str_length--;
  }
  return ECMD_FINAL(str_length);
}


int16_t
parse_cmd_sd_cd (char *cmd, char *output, uint16_t len)
{
  char tmp_path[MAX_CURRENT_PATH_LEN+1];
  char *ptr1, *ptr2;
  int tmp_path_len;

  while (*cmd == ' ')
    cmd ++;

  if (*cmd == 0) {
    cmd[0] = '/';
    cmd[1] = 0;
  }
  if (*cmd == '/') {
    strncpy(tmp_path, cmd, MAX_CURRENT_PATH_LEN);
  }
  else {
    strncpy(tmp_path, current_path, MAX_CURRENT_PATH_LEN);
    strncat(tmp_path, cmd, MAX_CURRENT_PATH_LEN - strlen(tmp_path));
  }
  tmp_path_len = strlen(tmp_path);
  if (tmp_path[tmp_path_len - 1] != '/') {
    tmp_path[tmp_path_len] = '/';
    tmp_path[tmp_path_len + 1] = 0;
    tmp_path_len++;
  }

  while ((ptr1 = strstr_P(tmp_path, PSTR("/../")))) {
     *ptr1 = 0;
     ptr2 = strrchr(tmp_path,'/');
     if (ptr2 == NULL) ptr2 = tmp_path;
     memmove(ptr2, ptr1+3, strlen(ptr1+3)+1); 
  }

  struct fat_dir_struct *fat_dir = vfs_sd_chdir(tmp_path);
  if (fat_dir) {
    if (wd) fat_close_dir(wd);
    wd = fat_dir;
    strncpy(current_path, tmp_path, MAX_CURRENT_PATH_LEN);
    return ECMD_FINAL_OK;
  }
  return ECMD_FINAL(snprintf_P(output, len, PSTR("No such directory.")));
}


int16_t
parse_cmd_sd_cat (char *cmd, char *output, uint16_t len)
{
  int read;
  char *filename;

  if (cmd[0] != 5) {
    filename = cmd;
    while (*filename == ' ')
      filename ++;
    fh = open_file_in_current_dir(wd, filename);
    if (fh == NULL)
      return ECMD_FINAL(snprintf_P(output, len, PSTR("No such file.")));
    cmd[0] = 5;
  }
  if ((read = fat_read_file(fh, (uint8_t*) output, len))) {
    output[read] = ECMD_NO_NEWLINE;
    return ECMD_AGAIN(read);
  }
  fat_close_file(fh);
  return ECMD_FINAL_OK;
}


int16_t
parse_cmd_sd_rm (char *cmd, char *output, uint16_t len)
{
  while (*cmd == ' ')
    cmd ++;

  struct fat_dir_entry_struct file_entry;
  if (find_file_in_current_dir(wd, cmd, &file_entry)) {
    if (!fat_delete_file(fat_get_fs(), &file_entry))
      return ECMD_FINAL(snprintf_P(output, len, PSTR("Error deleting file: %s"), cmd));
    return ECMD_FINAL_OK;
  }
  return ECMD_FINAL(snprintf_P(output, len, PSTR("No such file.")));
}


int16_t
parse_cmd_sd_info(char *cmd, char *output, uint16_t len)
{
  struct sd_raw_info disk_info;

  if (!fat_get_fs())
    return ECMD_FINAL(snprintf_P(output, len, PSTR("No filesystem.")));
  if (!sd_raw_get_info(&disk_info))
    return ECMD_FINAL(snprintf_P(output, len, PSTR("No disk info!")));
  return ECMD_FINAL(snprintf_P(output, len, PSTR("\n"              \
                                                 "manuf:  0x%02X\n"  \
                                                 "oem:    %s\n"    \
                                                 "prod:   %s\n"    \
                                                 "ser:    0x%lX\n" \
                                                 "total:  %lu MB\n"\
                                                 "size:   %lu\n"   \
                                                 "free:   %lu\n"), \
                                                 disk_info.manufacturer, \
                                                 disk_info.oem, \
                                                 disk_info.product, \
                                                 disk_info.serial, \
                                                 disk_info.capacity / 1024 / 1024, \
                                                 fat_get_fs_size(fat_get_fs()), \
                                                 fat_get_fs_free(fat_get_fs())));
}

#if 0
    uart_puts_p(PSTR("manuf:  0x")); uart_putc_hex(disk_info.manufacturer); uart_putc('\n');
    uart_puts_p(PSTR("oem:    ")); uart_puts((char*) disk_info.oem); uart_putc('\n');
    uart_puts_p(PSTR("prod:   ")); uart_puts((char*) disk_info.product); uart_putc('\n');
    uart_puts_p(PSTR("rev:    ")); uart_putc_hex(disk_info.revision); uart_putc('\n');
    uart_puts_p(PSTR("serial: 0x")); uart_putdw_hex(disk_info.serial); uart_putc('\n');
    uart_puts_p(PSTR("date:   ")); uart_putw_dec(disk_info.manufacturing_month); uart_putc('/');
                                   uart_putw_dec(disk_info.manufacturing_year); uart_putc('\n');
    uart_puts_p(PSTR("size:   ")); uart_putdw_dec(disk_info.capacity / 1024 / 1024); uart_puts_p(PSTR("MB\n"));
    uart_puts_p(PSTR("copy:   ")); uart_putw_dec(disk_info.flag_copy); uart_putc('\n');
    uart_puts_p(PSTR("wr.pr.: ")); uart_putw_dec(disk_info.flag_write_protect_temp); uart_putc('/');
                                   uart_putw_dec(disk_info.flag_write_protect); uart_putc('\n');
    uart_puts_p(PSTR("format: ")); uart_putw_dec(disk_info.format); uart_putc('\n');
    uart_puts_p(PSTR("free:   ")); uart_putdw_dec(fat_get_fs_free(fs)); uart_putc('/');
                                   uart_putdw_dec(fat_get_fs_size(fs)); uart_putc('\n');

    return 1;
}
#endif


/*
  -- Ethersex META --
  block([[SD-Karte]])
  ecmd_feature(sd_dir, "ls",, List contents of current SD directory.)
  ecmd_feature(sd_mkdir, "mkdir",PATH, Create directory hierarchy PATH.)
  ecmd_feature(sd_cd, "cd",PATH, Change directory PATH.)
  ecmd_feature(sd_cat, "cat",FILE, Print File in current directory PATH.)
  ecmd_feature(sd_pwd, "pwd",, Print current directory PATH.)
  ecmd_feature(sd_pup, "pup",, Print current upload PATH.)
  ecmd_feature(sd_rm, "rm",FILE, Delete File in current directory PATH.)
  ecmd_feature(sd_info, "sd info",, Print disk info.)
*/
